// CS 106X Fall 2018 - Nick Troccoli
// C++ solutions for problems from Section #2 (Week 3)
// Section topic: Introduction to Recursion

#include <iostream>
#include <fstream>
#include "console.h"
#include "stack.h"
#include "map.h"
#include "hashmap.h"
#include "vector.h"

using namespace std;

/**************************************************
 * 1. Recursion Code Trace - (recursion, tracing) *
 **************************************************/
// a) 4
// b) 16, 8, 16
// c) 12, 9 6, 3, 6, 9, 12
void mysteryTrace(int x, int y) {
    if (y == 1) {
        cout << x;
    } else {
        cout << (x*y) << ", ";
        mysteryTrace(x, y-1);
        cout << ", " << (x*y);
    }
}


/***************************************************
 * 2. Debugging Recursion - (recursion, debugging) *
 ***************************************************/
int recursiveMax(const Vector<int>& v, int left, int right) {
    if (left == right) {
        return v[left];
    } else if (left < right) {
        int middle = (left + right) / 2;
        int leftMax = recursiveMax(v, left, middle);
        int rightMax = recursiveMax(v, middle + 1, right); // changed middle -> middle + 1
        if (leftMax > rightMax) {
            return leftMax;
        } else {
            return rightMax;
        }
    } else {
        throw "Invalid range.";
    }
}



/********************************
 * 3. cannonballs - (recursion) *
 ********************************/
int cannonballs(int height) {
    if (height < 0) {
        throw height;  //can also use the Stanford library error function to add a message
    } else if (height == 0) {
        return 0;
    } else {
        return height * height + cannonballs(height - 1);
    }
}


/*******************************************
 * 4. reverseString - (recursion, strings) *
 *******************************************/
string reverseString(const string& s) {
    if (s == "") {
        return "";
    } else {
        return reverseString(s.substr(1)) + s[0];
    }
}


/***************************************
 * 5. doubleStack - (recursion, Stack) *
 ***************************************/
void doubleStack(Stack<int>& s) {
    if (!s.isEmpty()) {
        int next = s.pop();
        doubleStack(s);
        s.push(next);
        s.push(next);
    }
}


/*********************************
 * 6. combinations - (recursion) *
 *********************************/

// non-memoized solution
int combinations(int n, int k) {
    if (n <= 0 || k < 0 || k > n) {
        return 0;
    } else if (k == 0 || k == n) {
        return 1;
    } else {
        return combinations(n-1, k-1) + combinations(n-1, k);
    }
}

// bonus: memoized cool solution
int combinationsHelper(int n, int k, HashMap<int, HashMap<int, int>>& cache) {
    if (n <= 0 || k < 0 || k > n) {
        return 0;
    } else if (k == 0 || k == n) {
        return 1;
    } else if (cache.containsKey(n) && cache[n].containsKey(k)) {
        return cache[n][k];
    } else {
        int result = combinationsHelper(n-1, k-1, cache) + combinationsHelper(n-1, k, cache);
        cache[n][k] = result;
        return result;
    }
}

int combinationsBonus(int n, int k) {
    HashMap<int, HashMap<int, int>> cache;
    return combinationsHelper(n, k, cache);
}


/********************************
 * 7. isSubsequence (recursion) *
 ********************************/

//logic: compare letters until you find every letter in small or exhaust the letters in big
bool isSubsequence(const string& big, const string& small) {
    if (small == "") {
        return true;
    } else if (big == "") {
        return false;
    } else {
        if (big[0] == small[0]) {
            return isSubsequence(big.substr(1), small.substr(1));
        } else {
            return isSubsequence(big.substr(1), small);
        }
    }
}


/*************************
 * 8. reverseMap - (Map) *
 *************************/
Map<string, int> reverseMap(const Map<int, string>& map) {
    Map<string, int> rev;
    for (int key : map) {
        rev[map[key]] = key;
    }
    return rev;
}


/* main function for testing individual problems */

int main() {
    cout << "Problem 1:" << endl;
    mysteryTrace(4,1);
    cout << endl;
    mysteryTrace(8,2);
    cout << endl;
    mysteryTrace(3,4);
    cout << endl;

    cout << endl << "Problem 2:" << endl;
    Vector<int> vec{1, 2, 4, 2, 3, 5};
    cout << recursiveMax(vec, 0, 2) << endl;

    cout << endl << "Problem 3:" << endl;
    cout << cannonballs(3) << endl;

    cout << endl << "Problem 4:" << endl;
    cout << reverseString("Hi, there!") << endl;

    cout << endl << "Problem 5:" << endl;
    Stack<int> s{1, 2, 3};
    doubleStack(s);
    cout << s << endl;

    cout << endl << "Problem 6:" << endl;
    cout << combinations(36, 25) << endl;      // non-memoized
    cout << combinationsBonus(36, 25) << endl; // memoized; to demonstrate speed difference

    cout << endl << "Problem 7:" << endl;
    cout << isSubsequence("computer", "core") << endl;
    cout << isSubsequence("computer", "cope") << endl;
    cout << isSubsequence("computer", "computer") << endl;

    cout << endl << "Problem 8:" << endl;
    Map<int, string> map{{1,"a"},{2,"b"},{3,"c"}};
    cout << reverseMap(map) << endl;

    // you may do any other testing you want here.

    return 0;
}
